from itertools import islice
from typing import Any, Generator, KeysView, Mapping, Sequence, TypeVar, ValuesView


__all__ = [
    "batched",
]


T = TypeVar("T", Sequence, Mapping, KeysView, ValuesView)


def batched(iterable: T, batch_size: int) -> Generator[T, Any, None]:
    if batch_size <= 0:
        raise ValueError("Batch size must be positive.")

    while iterable:
        try:
            iterable, batch = iterable[batch_size:], iterable[:batch_size]

        except TypeError:
            if isinstance(iterable, Mapping):
                try:
                    iterable, batch = (
                        iterable.__class__(islice(iterable.items(), batch_size, len(iterable))),  # type: ignore
                        iterable.__class__(islice(iterable.items(), 0, batch_size)),  # type: ignore
                    )
                except TypeError:
                    iterable, batch = (
                        iterable.__class__(**dict(islice(iterable.items(), batch_size, len(iterable)))),  # type: ignore
                        iterable.__class__(**dict(islice(iterable.items(), 0, batch_size))),  # type: ignore
                    )
            else:
                try:
                    iterable, batch = (
                        iterable.__class__(islice(iter(iterable), batch_size, len(iterable))),
                        iterable.__class__(islice(iter(iterable), 0, batch_size)),
                    )
                except TypeError:
                    iterable, batch = (
                        list(islice(iter(iterable), batch_size, len(iterable))),
                        list(islice(iter(iterable), 0, batch_size)),
                    )

        yield batch



from typing import Any, Dict, Hashable, Iterator, KeysView, Tuple, ValuesView

from ._iterable_cache import iterable_cache


__all__ = [
    "Definition",
]


class Sentinel:
    pass


_ignore = {
    "__module__",
    "__qualname__",
    "__doc__",
    "__annotations__",
    "__new__",
    "__classcell__",
}

_internal = {
    "_key2value_map_",
    "_value2key_map_",
    "_unique_",
    "_default_key_",
    "_default_value_",
}


class DefinitionMeta(type):
    def __new__(
        cls,
        clss: str,
        bases: Tuple[type],
        classdict: Dict[str, Any],
        *,
        default_key: str = Sentinel,
        default_value: Any = Sentinel,
        unique: bool = True,
    ):
        _key_to_baseclass_: Dict[str, type] = {}
        _key2value_map_: Dict[str, Any] = {}
        _value2key_map_: Dict[Any, str] = {}
        for base in bases:
            for key, value in getattr(base, "_key2value_map_", {}).items():

                if unique and not isinstance(value, Hashable):
                    raise ValueError(
                        f"'{value}' is not hashable, and thus cannot be used for uniquely-valued Definitions."
                    )

                if key in _key2value_map_:
                    raise ValueError(f"'{_key_to_baseclass_[key].__name__}' and '{base.__name__}' both defined '{key}'")

                if unique and value in _value2key_map_:
                    raise ValueError(f"'{_value2key_map_[value]}' and '{key}' have the same value: {value}")

                _key_to_baseclass_[key] = base
                _key2value_map_[key] = value
                if unique:
                    _value2key_map_[value] = key

        _new_class_ = super().__new__(cls, clss, bases, classdict)
        _new_class_._key2value_map_: Dict[str, Any] = _key2value_map_  # type: ignore
        _new_class_._value2key_map_: Dict[Any, str] = _value2key_map_  # type: ignore
        _new_class_._unique_: bool = unique  # type: ignore
        _new_class_._default_key_: str = default_key  # type: ignore
        _new_class_._default_value_: Any = default_value  # type: ignore

        for key, value in classdict.items():
            if key in _ignore:
                continue

            if unique and not isinstance(value, Hashable):
                raise ValueError(f"'{value}' is not hashable, and thus cannot be used for uniquely-valued Definitions.")

            if key in _new_class_._key2value_map_:
                raise ValueError(f"'{_key_to_baseclass_[key].__name__}' already defined '{key}'")

            if unique and value in _new_class_._value2key_map_:
                raise ValueError(f"'{_new_class_._value2key_map_[value]}' and '{key}' have the same value: {value}")

            _new_class_._key2value_map_[key] = value
            if unique:
                _new_class_._value2key_map_[value] = key

        return _new_class_

    def __setattr__(cls, key: Any, value: Any) -> None:
        # Allow specific attributes to be set once
        if key in _internal and key not in cls.__dict__:
            return super().__setattr__(key, value)

        raise AttributeError("Definition cannot be changed once created.")

    def __delattr__(cls, key: Any) -> None:
        raise AttributeError("Definition cannot be changed once created.")

    def __getitem__(cls, key: Any) -> Any:
        try:
            return cls._key2value_map_[key]
        except KeyError as error:
            if cls._default_value_ is not Sentinel:
                return cls._default_value_
            raise error

    def __call__(cls, value: Any = Sentinel) -> str:  # type: ignore
        if value is Sentinel:
            raise TypeError("Definition should not be instantiated.")

        if not cls._unique_:
            raise TypeError("Non unique-valued Definitions do not support reverse lookups.")

        try:
            return cls._value2key_map_[value]
        except KeyError as error:
            if cls._default_key_ is not Sentinel:
                return cls._default_key_
            raise error

    def __contains__(cls, item: Any) -> bool:
        return item in cls.values()  # pylint: disable=no-value-for-parameter

    def __str__(cls) -> str:
        return "{" + ", ".join([f"{key}={value}" for key, value in cls._key2value_map_.items()]) + "}"

    def __len__(cls) -> int:
        return len(cls._key2value_map_)

    def __iter__(cls) -> "DefinitionMeta":  # pylint: disable=non-iterator-returned
        return cls

    @iterable_cache(provider="values")
    def __next__(cls, keys: Iterator[str]) -> str:
        return next(keys)

    def keys(cls) -> KeysView[str]:
        """All definition keys in the Definition."""
        return cls._key2value_map_.keys()

    def values(cls) -> ValuesView[Any]:
        """All definition values in the Definition."""
        return cls._key2value_map_.values()


class Definition(metaclass=DefinitionMeta):
    r"""Create Definitions by subclassing this class.

    You can use the following class-level keywork arguments
    to change how the Definition behaves:

                  Arguments go here↴
    class MyDefinition(Definition, key=...):
        ...

    default_key: str — When no key is for a given value is found,
                       return this instead of raising a KeyError.
    default_value: Any — When no value is for a given key is found,
                         return this instead of raising a KeyError.
    unique: bool — By default, all keys and values in a definition have
                   to be unique. You can set unique=False to disable this,
                   but it also disables reverse lookup with values.
    """



from ._batched import batched
from ._definition import Definition
from ._import_from_string import import_from_string
from ._iterable_cache import iterable_cache


__all__ = [
    "batched",
    "Definition",
    "import_from_string",
    "iterable_cache",
]



from importlib import import_module
from sys import modules
from typing import Any


__all__ = [
    "import_from_string",
]


def import_from_string(string: str) -> Any:
    """Attempt to import from a dot import string representation.

    :param string: String to attempt to import with.
    :raises ImportError: Could not import from string.
    """

    msg = f"Could not import {string!r}"
    try:
        module_path, class_name = string.rsplit(".", 1)
    except ValueError as error:
        raise ImportError(f"{msg}: Not a valid module path.") from error

    if module_path not in modules or (
        # Module is not fully initialized.
        getattr(modules[module_path], "__spec__", None) is not None
        and getattr(modules[module_path].__spec__, "_initializing", False) is True
    ):
        try:
            import_module(module_path)
        except ImportError as error:
            raise ImportError(f"{msg}: {error}.") from error

    try:
        return getattr(modules[module_path], class_name)
    except AttributeError as error:
        raise ImportError(f"{msg}: Module {module_path!r} does not define {class_name!r}.") from error



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['magic_specs']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'magic-specs',
    'version': '0.0.6',
    'description': '',
    'long_description': '# Magic value specification utilities\n\n[![Coverage Status][coverage-badge]][coverage]\n[![GitHub Workflow Status][status-badge]][status]\n[![PyPI][pypi-badge]][pypi]\n[![GitHub][licence-badge]][licence]\n[![GitHub Last Commit][repo-badge]][repo]\n[![GitHub Issues][issues-badge]][issues]\n[![Python Version][version-badge]][pypi]\n\n```shell\npip install magic-specs\n```\n\n---\n\n**Documentation**: [https://mrthearman.github.io/magic-specs/](https://mrthearman.github.io/magic-specs/)\n\n**Source Code**: [https://github.com/MrThearMan/magic-specs/](https://github.com/MrThearMan/magic-specs/)\n\n---\n\n\nThis library contains utilities for making definitions for magic values.\n\n\n[coverage-badge]: https://coveralls.io/repos/github/MrThearMan/magic-specs/badge.svg?branch=main\n[status-badge]: https://img.shields.io/github/workflow/status/MrThearMan/magic-specs/Tests\n[pypi-badge]: https://img.shields.io/pypi/v/magic-specs\n[licence-badge]: https://img.shields.io/github/license/MrThearMan/magic-specs\n[repo-badge]: https://img.shields.io/github/last-commit/MrThearMan/magic-specs\n[issues-badge]: https://img.shields.io/github/issues-raw/MrThearMan/magic-specs\n[version-badge]: https://img.shields.io/pypi/pyversions/magic-specs\n\n[coverage]: https://coveralls.io/github/MrThearMan/magic-specs?branch=main\n[status]: https://github.com/MrThearMan/magic-specs/actions/workflows/main.yml\n[pypi]: https://pypi.org/project/magic-specs\n[licence]: https://github.com/MrThearMan/magic-specs/blob/main/LICENSE\n[repo]: https://github.com/MrThearMan/magic-specs/commits/main\n[issues]: https://github.com/MrThearMan/magic-specs/issues',
    'author': 'Matti Lamppu',
    'author_email': 'lamppu.matti.akseli@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/MrThearMan/magic-specs',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



from functools import wraps
from typing import Any, Callable


__all__ = [
    "iterable_cache",
]


def iterable_cache(provider: str, *args, **kwargs) -> Callable[..., Any]:
    r"""Decorate classes "__next__" magic method. This creates a cache that remembers
    the current iterator object the instance is iterating over. The cached instance is
    provided to "__next__" as an argument, so it can just call next(...) on it.
    Once the iterator is iterated, it clears the cached iterator,
    so the class can then be iterated again.

    class A:
        def __iter__(self):
            return self

        @iterable_cache(provider="keys")
        def __next__(self, keys):
            return

        def keys(self):
            # items are provided here

    :params provider: Function that procides the items to iterate over.
                      Additional arguments can be provided to it from *args and **kwargs.
    """

    def decorator(func) -> Callable:
        items = None

        @wraps(func)
        def wrapper(self) -> Any:
            nonlocal items
            if items is None:
                items = iter(getattr(self, provider)(*args, **kwargs))

            try:
                result = func(self, items)
                return result
            except StopIteration as error:
                items = None
                raise error

        return wrapper

    return decorator
